import "webclgl";
import "nclgl";
import {Component} from "./Component.class";
import {Constants} from "./Constants";

/**
 * @class
 * @constructor
 */
export class Component_GPU extends Component {
    constructor() {
        super();

        this.type = Constants.COMPONENT_TYPES.GPU;
        this.node = null;
        this.gl = null;

        this.gpufG = null;
        this.args = {};

        this.onPreProcessKernel_fn = null;
        this.onPostProcessKernel_fn = null;
        this.onPreProcessGraphic_fn = null;
        this.onPostProcessGraphic_fn = null;
    }


    /**
     * initialize
     * @param {Node} nod
     * @param {WebGLRenderingContext} glCtx.
     * @override
     */
    initialize(nod, glCtx) {
        this.node = nod;
        this.gl = glCtx;
    };

    /**
     * tick
     * @override
     */
    tick(delta) {
        console.log("\n[COMP_GPU] tick arguments");
        this.tickArguments();

        if(this.onPreProcessKernel_fn != null)
            this.onPreProcessKernel_fn();

        if(this.gpufG != null)
            this.gpufG.processKernels();

        if(this.onPostProcessKernel_fn != null)
            this.onPostProcessKernel_fn();


        if(this.onPreProcessGraphic_fn != null)
            this.onPreProcessGraphic_fn();

        if(this.gpufG != null)
            (typeof Module === "undefined") ? this.gpufG.processGraphic(undefined) : this.gpufG.processGraphic("null");

        if(this.onPostProcessGraphic_fn != null)
            this.onPostProcessGraphic_fn();
    };

    /**
     * setGPUFor
     */
    setGPUFor() {
        for(let key in arguments[1]) {
            let expl = key.split(" ");
            if(expl != null && expl.length > 1) {
                let argName = expl[1];
                this.args[argName] = {
                	"fnvalue": arguments[1][key],
                    "updatable": null,
                    "overrideDimensions": null};
            }
            arguments[1][key] = arguments[1][key]();
        }

        let onl = arguments[arguments.length-1];
        if(typeof Module === "undefined") {
            let F = () => {
                return gpufor.apply(this, arguments);
            };
            this.gpufG = new F();
            onl();
        } else {
            arguments[arguments.length-1] = (function(onl, gpufG) {
                this.gpufG = gpufG;
                onl();
            }).bind(this, onl);

            gpufor_asm.apply(this, arguments);
        }
    };

    onPreProcessKernel(fn) {
        this.onPreProcessKernel_fn = fn;
    }

    onPostProcessKernel(fn) {
        this.onPostProcessKernel_fn = fn;
    }

    onPreProcessGraphic(fn) {
        this.onPreProcessGraphic_fn = fn;
    }

    onPostProcessGraphic(fn) {
        this.onPostProcessGraphic_fn = fn;
    }

    destroyGPUFor() {
        if(typeof Module !== "undefined")
            this.gpufG.destroyContext();
    }

    /**
     * getWebCLGL
     * @returns {WebCLGL}
     */
    getWebCLGL() {
        return this.gpufG.getWebCLGL();
    };


    // █████╗ ██████╗  ██████╗ ██╗   ██╗███╗   ███╗███████╗███╗   ██╗████████╗███████╗
    //██╔══██╗██╔══██╗██╔════╝ ██║   ██║████╗ ████║██╔════╝████╗  ██║╚══██╔══╝██╔════╝
    //███████║██████╔╝██║  ███╗██║   ██║██╔████╔██║█████╗  ██╔██╗ ██║   ██║   ███████╗
    //██╔══██║██╔══██╗██║   ██║██║   ██║██║╚██╔╝██║██╔══╝  ██║╚██╗██║   ██║   ╚════██║
    //██║  ██║██║  ██║╚██████╔╝╚██████╔╝██║ ╚═╝ ██║███████╗██║ ╚████║   ██║   ███████║
    //╚═╝  ╚═╝╚═╝  ╚═╝ ╚═════╝  ╚═════╝ ╚═╝     ╚═╝╚══════╝╚═╝  ╚═══╝   ╚═╝   ╚══════╝

    /**
     *
     */
    addArgument(arg, value) {
        this.args[arg.split(" ")[1]] = {
            "fnvalue": value,
            "updatable": null,
            "overrideDimensions": null};

        this.gpufG.addArg(arg);
    };

    /**
     * @param {String} argument Argument to set
     * @param {Function} fnvalue
     * @returns {WebCLGLBuffer}
     */
    setArg(argument, fnvalue) {
        let val = fnvalue();
        let ovDim = null;
        let buff = null;

        if(val instanceof HTMLImageElement === true && typeof Module !== "undefined") {
            let e = document.createElement('canvas');
            e.width = val.width;
            e.height = val.height;

            ovDim = [val.width, val.height];

            let ctx2D_tex = e.getContext("2d");
            ctx2D_tex.drawImage(val, 0, 0);
            let arrayTex = ctx2D_tex.getImageData(0, 0, val.width, val.height);

            let arN = new Float32Array(arrayTex.data.length);
            for(let n=0; n < arrayTex.data.length; n++)
                arN[n] = arrayTex.data[n]/256;

            val = arN;

            buff = this.gpufG.setArg(argument, val, ovDim);
            this.args[argument] = {	"fnvalue": function() {return val;},
                "updatable": null,
                "overrideDimensions": ovDim};
        } else {
            buff = this.gpufG.setArg(argument, fnvalue(), null);
            this.args[argument] = {	"fnvalue": fnvalue,
                "updatable": null,
                "overrideDimensions": null};
        }

        return buff;
    };

    /**
     * getComponentBufferArg
     * @param {String} argument Argument to set
     * @param {Component|Component_GPU} comp_gpu
     */
    getComponentBufferArg(argument, comp_gpu) {
        this.gpufG.getGPUForArg(argument, comp_gpu.gpufG);
        this.args[argument] = {	"fnvalue": null,
            "updatable": null,
            "overrideDimensions": null};
    };

    /**
     * getArgs
     * @returns {Object}
     */
    getArgs() {
        return this.args;
    };

    /**
     * getAllArgs
     * @returns {Object}
     */
    getAllArgs() {
        return this.gpufG.getAllArgs();
    };

    /**
     * getBuffers
     * @returns {{}|Array<WebCLGLBuffer>}
     */
    getBuffers() {
        return this.gpufG._argsValues;
    };

    argIsNull(argument) {
        return this.gpufG.argIsNull(argument);
    }

    getArgBufferWidth(argument) {
        return this.gpufG.getArgBufferWidth(argument);
    }

    /**
     * @param {String} argument Argument to set
     * @param {boolean} value
     */
    setArgUpdatable(argument, value) {
        this.args[argument].updatable = value;
    };

    /**
     * tickArguments
     */
    tickArguments() {
        for(let key in this.args) {
            if(this.args[key].updatable === true) {
                let arg = this.args[key];
                this.gpufG.setArg(key, arg.fnvalue(), arg.overrideDimensions);
            }
        }
    };

}
global.Component_GPU = Component_GPU;
module.exports.Component_GPU = Component_GPU;